import pytest

from sktmls.models import MLSModelError
from sktmls.models.contrib import MbrValidCategoryBaselineModel


class TestMbrValidCategoryBaselineModel:
    def test_000_mbr_valid_baseline_1(self):
        mbm = MbrValidCategoryBaselineModel(
            model_name="test_model",
            model_version="test_version",
            features=[
                "user_id",
                "available_benefit_activity",
                "available_benefit_asianfood",
                "available_benefit_beauty",
            ],
            categories=["mbr_activity", "mbr_asianfood", "mbr_beauty"],
            item_type="test_type",
            num_pick=1,
        )

        results = mbm.predict(["1234567890", ["a1", "a2", "a3", "a4"], None, ["b1", "b2", "3", "4"]])
        assert len(results["items"]) == 1

        item = results["items"][0]
        assert item["id"] == "test_type"
        assert item["type"] == "test_type"
        assert len(item["props"]["test_type"]) == 1
        assert "mbr_activity" in item["props"]["test_type"] or "mbr_beauty" in item["props"]["test_type"]

    def test_001_mbr_valid_baseline_2(self):
        mbm = MbrValidCategoryBaselineModel(
            model_name="test_model",
            model_version="test_version",
            features=[
                "user_id",
                "available_benefit_activity",
                "available_benefit_asianfood",
                "available_benefit_beauty",
            ],
            categories=["mbr_activity", "mbr_asianfood", "mbr_beauty"],
            item_type="test_type",
            num_pick=2,
        )

        results = mbm.predict(["1234567890", ["a1", "a2", "a3", "a4"], None, ["b1", "b2", "3", "4"]])
        assert len(results["items"]) == 1

        item = results["items"][0]
        assert item["id"] == "test_type"
        assert item["type"] == "test_type"
        assert len(item["props"]["test_type"]) == 2
        assert "mbr_activity" in item["props"]["test_type"] and "mbr_beauty" in item["props"]["test_type"]

    def test_002_mbr_valid_baseline_2_many_none(self):
        mbm = MbrValidCategoryBaselineModel(
            model_name="test_model",
            model_version="test_version",
            features=[
                "user_id",
                "available_benefit_activity",
                "available_benefit_asianfood",
                "available_benefit_beauty",
            ],
            categories=["mbr_activity", "mbr_asianfood", "mbr_beauty"],
            item_type="test_type",
            num_pick=2,
        )

        results = mbm.predict(["1234567890", ["a1", "a2", "a3", "a4"], None, None])
        assert len(results["items"]) == 1

        item = results["items"][0]
        assert item["id"] == "test_type"
        assert item["type"] == "test_type"
        assert len(item["props"]["test_type"]) == 1
        assert "mbr_activity" in item["props"]["test_type"]

    def test_003_mbr_valid_baseline_2_all_none(self):
        mbm = MbrValidCategoryBaselineModel(
            model_name="test_model",
            model_version="test_version",
            features=[
                "user_id",
                "available_benefit_activity",
                "available_benefit_asianfood",
                "available_benefit_beauty",
            ],
            categories=["mbr_activity", "mbr_asianfood", "mbr_beauty"],
            item_type="test_type",
            num_pick=1,
        )

        results = mbm.predict(["1234567890", ["a1"], None, None])
        assert len(results["items"]) == 0

    def test_004_mbr_valid_baseline_invalid(self):
        mbm = MbrValidCategoryBaselineModel(
            model_name="test_model",
            model_version="test_version",
            features=[
                "user_id",
                "available_benefit_activity",
                "available_benefit_asianfood",
                "available_benefit_beauty",
            ],
            categories=["mbr_activity", "mbr_asianfood", "mbr_beauty"],
            item_type="test_type",
            num_pick=1,
        )

        with pytest.raises(MLSModelError):
            mbm.predict(["1234567890", ["a1", "a2", "a3", "a4"], None, "good"])



import pytest

from sktmls.models.contrib import MbrVipInfoRuleModel


class TestMbrVipInfoRuleModels:
    @pytest.fixture(scope="class")
    def param(self):
        return {
            "model": MbrVipInfoRuleModel("test_model", "test_version", ["mbr_card_gr_cd", "mbr_use_discount_amt_cum"])
        }

    def test_000_mbr_vip_info_return(self, param):
        assert param.get("model").predict(["N", 1]) == {
            "items": [{"id": "INF0000009", "name": "VIP 혜택 및 승급 방법 안내", "type": "mbr_vip_info", "props": {}}]
        }

    def test_001_mbr_vip_info_no_return(self, param):
        assert param.get("model").predict(["V", 1]) == {"items": []}

    def test_002_mbr_vip_info_no_return(self, param):
        assert param.get("model").predict(["N", 0]) == {"items": []}



import mxnet as mx
import logging

logger = logging.getLogger(__name__)

__all__ = ['AddJitterOp', 'AddJitterOpProp']


INITIAL_JITTER_FACTOR = 1e-9
JITTER_GROWTH = 10.
JITTER_UPPERBOUND_FACTOR = 1e3


class AddJitterOp(mx.operator.CustomOp):
    """
    Finds smaller jitter to add to diagonal of square matrix to render the
    matrix positive definite (in that linalg.potrf works).

    Given input x (positive semi-definite matrix) and sigsq_init (nonneg
    scalar), find sigsq_final (nonneg scalar), so that:
        sigsq_final = sigsq_init + jitter, jitter >= 0,
        x + sigsq_final * Id positive definite (so that potrf call works)
    We return the matrix x + sigsq_final * Id, for which potrf has not failed.

    For the gradient, the dependence of jitter on the inputs is ignored.

    The values tried for sigsq_final are:
        sigsq_init, sigsq_init + initial_jitter * (jitter_growth ** k),
        k = 0, 1, 2, ...,
        initial_jitter = initial_jitter_factor * mean(diag(x))

    Note: The scaling of initial_jitter with mean(diag(x)) is taken from GPy.
    The rationale is that the largest eigenvalue of x is >= mean(diag(x)), and
    likely of this magnitude.

    There is no guarantee that the Cholesky factor returned is well-conditioned
    enough for subsequent computations to be reliable. A better solution
    would be to estimate the condition number of the Cholesky factor, and to add
    jitter until this is bounded below a threshold we tolerate. See

        Higham, N.
        A Survey of Condition Number Estimation for Triangular Matrices
        MIMS EPrint: 2007.10

    Algorithm 4.1 could work for us.
    """

    def __init__(
            self, initial_jitter_factor, jitter_growth, debug_log, **kwargs):
        super(AddJitterOp, self).__init__(**kwargs)

        assert initial_jitter_factor > 0. and jitter_growth > 1.
        self._initial_jitter_factor = initial_jitter_factor
        self._jitter_growth = jitter_growth
        self._debug_log = debug_log

    def _get_constant_identity(self, x, constant):
        n, _ = x.shape
        return mx.nd.diag(
            mx.nd.ones(shape=(n,), ctx=x.context, dtype=x.dtype) * constant)

    def _get_jitter_upperbound(self, x):
        # To define a safeguard in the while-loop of the forward,
        # we define an upperbound on the jitter we can reasonably add
        # the bound is quite generous, and is dependent on the scale of the input x
        # (the scale is captured via the trace of x)
        # the primary goal is avoid any infinite while-loop.
        return JITTER_UPPERBOUND_FACTOR * max(
            1., mx.nd.mean(mx.nd.diag(x)).asscalar())

    def forward(self, is_train, req, in_data, out_data, aux):
        x = in_data[0]
        sigsq_init = in_data[1]
        jitter = 0.
        jitter_upperbound = self._get_jitter_upperbound(x)
        must_increase_jitter = True
        x_plus_constant = None
        while must_increase_jitter and jitter <= jitter_upperbound:
            try:
                x_plus_constant = x + self._get_constant_identity(
                    x, sigsq_init + jitter)
                L = mx.nd.linalg.potrf(x_plus_constant)
                # because of the implicit asynchronous processing in MXNet,
                # we need to enforce the computation of L to happen right here
                L.wait_to_read()
                must_increase_jitter = False
            except mx.base.MXNetError:
                if self._debug_log == 'true':
                    logger.info("sigsq = {} does not work".format(
                        sigsq_init.asscalar() + jitter))
                if jitter == 0.0:
                    jitter = self._initial_jitter_factor * mx.nd.mean(
                        mx.nd.diag(x)).asscalar()
                else:
                    jitter *= self._jitter_growth

        assert not must_increase_jitter,\
            "The jitter ({}) has reached its upperbound ({}) while the Cholesky of the input matrix still cannot be computed.".format(jitter, jitter_upperbound)
        if self._debug_log == 'true':
            _sigsq_init = sigsq_init.asscalar()
            logger.info("sigsq_final = {}".format(_sigsq_init + jitter))
        self.assign(out_data[0], req[0], x_plus_constant)

    def backward(self, req, out_grad, in_data, out_data, in_grad, aux):
        self.assign(in_grad[0], req[0], out_grad[0])
        trace_out_grad = mx.nd.sum(mx.nd.diag(out_grad[0]))
        self.assign(in_grad[1], req[1], trace_out_grad)


@mx.operator.register("add_jitter")
class AddJitterOpProp(mx.operator.CustomOpProp):
    def __init__(
            self, initial_jitter_factor=INITIAL_JITTER_FACTOR,
            jitter_growth=JITTER_GROWTH, debug_log='false'):
        super(AddJitterOpProp, self).__init__(need_top_grad=True)
        # We need to cast the arguments
        # see detailed example https://github.com/Xilinx/mxnet/blob/master/docs/tutorials/gluon/customop.md
        self._initial_jitter_factor = float(initial_jitter_factor)
        self._jitter_growth = float(jitter_growth)
        self._debug_log = debug_log

    def list_arguments(self):
        return ['x', 'sigsq_init']

    def list_outputs(self):
        return ['x_plus_sigsq_final']

    def infer_shape(self, in_shape):
        x_shape = in_shape[0]
        assert len(x_shape) == 2 and x_shape[0] == x_shape[1], \
            "x must be square matrix, shape (n, n)"
        ssq_shape = in_shape[1]
        assert len(ssq_shape) == 1 and ssq_shape[0] == 1, \
            "sigsq_init must be scalar, shape (1,)"
        return in_shape, [x_shape], []

    def create_operator(self, ctx, shapes, dtypes, **kwargs):
        return AddJitterOp(
            initial_jitter_factor=self._initial_jitter_factor,
            jitter_growth=self._jitter_growth,
            debug_log=self._debug_log, **kwargs)



from typing import Set, Tuple, Dict
import logging
import numbers

logger = logging.getLogger(__name__)


class CheckType(object):
    def assert_valid(self, key: str, value):
        pass


class Float(CheckType):
    def __init__(self, lower: float = None, upper: float = None):
        if lower and upper:
            assert lower < upper
        self.lower = lower
        self.upper = upper

    def assert_valid(self, key: str, value):
        assert isinstance(value, numbers.Real), \
            "{}: Value = {} must be of type float".format(key, value)
        assert (not self.lower) or value >= self.lower, \
            "{}: Value = {} must be >= {}".format(key, value, self.lower)
        assert (not self.upper) or value <= self.upper, \
            "{}: Value = {} must be <= {}".format(key, value, self.upper)


class Integer(CheckType):
    def __init__(self, lower: int = None, upper: int = None):
        if lower and upper:
            assert lower < upper
        self.lower = lower
        self.upper = upper

    def assert_valid(self, key: str, value):
        assert isinstance(value, numbers.Integral), \
            "{}: Value = {} must be of type int".format(key, value)
        assert (not self.lower) or value >= self.lower, \
            "{}: Value = {} must be >= {}".format(key, value, self.lower)
        assert (not self.upper) or value <= self.upper, \
            "{}: Value = {} must be <= {}".format(key, value, self.upper)


class Categorical(CheckType):
    def __init__(self, choices: Tuple[str, ...]):
        self.choices = set(choices)

    def assert_valid(self, key: str, value):
        assert isinstance(value, str) and value in self.choices, \
            "{}: Value = {} must be in {}".format(key, value, self.choices)


class String(CheckType):
    def assert_valid(self, key: str, value):
        assert isinstance(value, str), "{}: Value = {} must be of type str"


class Boolean(CheckType):
    def assert_valid(self, key: str, value):
        assert isinstance(value, bool), \
            "{}: Value = {} must be boolean".format(key, value)


def check_and_merge_defaults(
        options: dict, mandatory: Set[str], default_options: dict,
        constraints: Dict[str, CheckType] = None, dict_name=None) -> dict:
    """
    First, check that all keys in mandatory appear in options. Second, create
    result_options by merging options and default_options, where entries in
    options have precedence. Finally, if constraints is given, this is used to
    check validity of values.

    :param options:
    :param mandatory:
    :param default_options:
    :param constraints:
    :param dict_name:
    :return: result_options
    """
    prefix = "" if dict_name is None else "{}: ".format(dict_name)
    for key in mandatory:
        assert key in options, \
            prefix + "Key '{}' is missing (but is mandatory)".format(key)
    log_msg = ""
    result_options = {
        k: v for k, v in options.items() if v is not None}
    for key, value in default_options.items():
        if key not in result_options:
            log_msg += (prefix + "Key '{}': Imputing default value {}\n".format(
                key, value))
            result_options[key] = value
    if log_msg:
        logger.info(log_msg)
    # Check constraints
    if constraints:
        for key, value in result_options.items():
            check = constraints.get(key)
            if check:
                check.assert_valid(prefix + "Key '{}'".format(key), value)

    return result_options


def filter_by_key(options: dict, remove_keys: Set[str]) -> dict:
    """
    Filter options by removing entries whose keys are in remove_keys.
    Used to filter kwargs passed to a constructor, before passing it to
    the superclass constructor.

    :param options:
    :param remove_keys:
    :return: Filtered options
    """
    return {
        k: v for k, v in options.items() if k not in remove_keys}


def assert_no_invalid_options(options: dict, all_keys: Set[str], name: str):
    for k in options:
        assert k in all_keys, \
            "{}: Invalid argument '{}'".format(name, k)



# This file contains various constants required for the definition of the model
# or to set up the optimization

from typing import NamedTuple
import numpy as np


DEFAULT_ENCODING = "logarithm"  # the other choices is positive

NUMERICAL_JITTER = 1e-9

INITIAL_NOISE_VARIANCE = 1e-3
INITIAL_MEAN_VALUE = 0.0
INITIAL_COVARIANCE_SCALE = 1.0
INITIAL_INVERSE_BANDWIDTHS = 1.0
INITIAL_WARPING = 1.0

INVERSE_BANDWIDTHS_LOWER_BOUND = 1e-4
INVERSE_BANDWIDTHS_UPPER_BOUND = 100

COVARIANCE_SCALE_LOWER_BOUND = 1e-3
COVARIANCE_SCALE_UPPER_BOUND = 1e3

NOISE_VARIANCE_LOWER_BOUND = 1e-9
NOISE_VARIANCE_UPPER_BOUND = 1e6

WARPING_LOWER_BOUND = 0.25
WARPING_UPPER_BOUND = 4.

MIN_POSTERIOR_VARIANCE = 1e-12

MIN_CHOLESKY_DIAGONAL_VALUE = 1e-10

DATA_TYPE = np.float64


class OptimizationConfig(NamedTuple):
    lbfgs_tol: float
    lbfgs_maxiter : int
    verbose : bool
    n_starts : int


class MCMCConfig(NamedTuple):
    n_samples : int
    n_burnin : int
    n_thinning : int


DEFAULT_OPTIMIZATION_CONFIG = OptimizationConfig(
    lbfgs_tol=1e-6,
    lbfgs_maxiter=500,
    verbose=False,
    n_starts=5)

DEFAULT_MCMC_CONFIG = MCMCConfig(
    n_samples=300,
    n_burnin=250,
    n_thinning=5)



from collections import OrderedDict
from datetime import datetime
from typing import Any, Dict, List

from pytz import timezone

from sktmls import MLSRuntimeENV
from sktmls.apis import MLSProfileAPIClient
from sktmls.models import MLSRuleModel, MLSModelError

TZ = timezone("Asia/Seoul")


class MbrPopularBenefitModel(MLSRuleModel):
    """
    MLS 모델 레지스트리에 등록되는 Rule 기반 클래스입니다.

    멤버십 인기 혜택 추천에 특화된 모델입니다.
    """

    def __init__(
        self,
        model_name: str,
        model_version: str,
        features: List[str],
        categories: List[str],
        item_type: str,
        hourly_contexts: List[Any],
        num_pick: int = 1,
    ):
        """
        ## Args

        - model_name: (str) 모델 이름
        - model_version: (str) 모델 버전
        - features: (list(str)) 피쳐 리스트 (필수 피쳐: `user_id`, `sex_cd`, `age`, `mbr_card_gr_cd`)
        - categories: (list(str)) 전체 카테고리 리스트. 필수 피쳐 이후 피쳐 순서와 동일해야 합니다.
        - item_type: (str) 결과의 `type`에 표기할 값
        - hourly_contexts: (list) context_id 생성을 위한 정보를 담은 시간별 context 리스트
        - num_pick: (optional) (int) 랜덤 추출할 갯수 (기본값: 1)

        ## Example

        ```python
        my_model_v1 = MbrPopularBenefitModel(
            model_name="my_model",
            model_version="v1",
            features=["user_id", "sex_cd", "age", "mbr_card_gr_cd", "available_benefit_activity", "available_benefit_asianfood", "available_benefit_beauty"],
            categories=["mbr_activity", "mbr_asianfood", "mbr_beauty"],
            item_type="mbr_popular_benefit",
            hourly_contexts=[30, 31,...],
            num_pick=14,
        )
        ```
        """
        assert (
            isinstance(features, list)
            and len(features) > 0
            and features[0] == "user_id"
            and features[1] == "sex_cd"
            and features[2] == "age"
            and features[3] == "mbr_card_gr_cd"
        ), "`features`가 유효하지 않습니다. 위 모델의 필수 피쳐는 `user_id`, `sex_cd`, `age`, `mbr_card_gr_cd`로 순서를 지켜 입력하여야합니다"
        assert isinstance(categories, list) and len(categories) + 4 == len(
            features
        ), "전체 카테고리 리스트 길이가 필수 피쳐를 제외한 피쳐의 수와 다릅니다."
        assert (
            isinstance(hourly_contexts, list) and len(hourly_contexts) == 24
        ), "`hourly_context`가 리스트가 아니거나 유효한 길이(24)가 아닙니다"
        assert isinstance(num_pick, int) and num_pick > 0, "`num_pick`이 유효하지 않습니다. 반드시 1 이상의 값을 가져야 합니다."

        super().__init__(model_name, model_version, features)

        self.categories = categories
        self.item_type = item_type
        self.hourly_contexts = hourly_contexts
        self.num_pick = num_pick

    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        for i, f in enumerate(x[4:]):
            if f is not None and not isinstance(f, list):
                raise MLSModelError(f"MbrPopularBenefitModel: {self.features[i + 1]}이 리스트 형식으로 전달되지 않았습니다.")

        pf_client = kwargs.get("pf_client") or MLSProfileAPIClient(runtime_env=MLSRuntimeENV.MMS)

        sex_cd = x[1]
        age = x[2]
        mbr_card_gr_cd = x[3]

        if (
            not isinstance(sex_cd, str)
            or not isinstance(age, int)
            or not isinstance(mbr_card_gr_cd, str)
            or sex_cd not in ["1", "2"]
            or age not in range(100)
            or mbr_card_gr_cd not in ["S", "G", "V"]
        ):
            return {"items": []}

        this_time = datetime.now(TZ)
        this_hour = this_time.strftime("%H")
        this_dow = str(this_time.weekday())

        available_benefit = {
            category: x[i + 4]
            for i, category in enumerate(self.categories)
            if isinstance(x[i + 4], list) and len(x[i + 4]) >= 1
        }

        if sex_cd == "1":
            seg_sex_cd = "male"
        else:
            seg_sex_cd = "female"

        if age in range(0, 19):
            seg_age = "age0019"
        elif age in range(19, 34):
            seg_age = "age2034"
        elif age in range(34, 49):
            seg_age = "age3549"
        else:
            seg_age = "age5099"

        seg_id = f"mbr_seg_{seg_sex_cd}_{seg_age}_gr_{mbr_card_gr_cd.lower()}"

        if this_dow in ["0", "1", "2", "3", "4"]:
            weekday_cd = "weekday"
        else:
            weekday_cd = "weekend"
        ranking_dimension = f"benefit_ranking_{weekday_cd}_hour_{this_hour}"

        item_profile = pf_client.get_item_profile(
            profile_id="tmembership_seg",
            item_id=seg_id,
            keys=[ranking_dimension],
        )

        if (
            ranking_dimension not in item_profile
            or not isinstance(item_profile[ranking_dimension], list)
            or not len(item_profile[ranking_dimension]) >= 1
        ):
            return {"items": []}

        target_benefit_list = item_profile[ranking_dimension]

        available_benefit_list = sum(available_benefit.values(), [])

        valid_target_benefit_list = list(
            OrderedDict.fromkeys(filter(lambda x: x in available_benefit_list, target_benefit_list))
        )

        candidates = [
            {
                "id": i,
                "name": "#",
                "type": self.item_type,
                "props": {"context_id": self.hourly_contexts[int(this_hour)]},
            }
            for i in valid_target_benefit_list
        ]
        num_pick = min(self.num_pick, len(candidates))
        top_candidates = candidates[:num_pick]

        return {"items": top_candidates}



import re
from abc import ABCMeta, abstractmethod
from typing import Any, Dict, List

import catboost
import lightgbm
import xgboost

from sktmls import ModelRegistry, MLSENV, MLSRuntimeENV, autogluon


class MLSModelError(Exception):
    def __init__(self, msg):
        super().__init__(msg)


class MLSModel(metaclass=ABCMeta):
    """
    MLS 모델 레지스트리에 등록되는 최상위 클래스입니다.
    """

    def __init__(self, model_name: str, model_version: str):
        assert type(model_name) == str
        assert type(model_version) == str

        if not bool(re.search("^[A-Za-z0-9_]+_model$", model_name)):
            raise MLSModelError(
                "model_name should follow naming rule. MUST be in alphabet, number, underscore and endwiths '_model'"
            )

        if not bool(re.search("^[A-Za-z0-9_]+$", model_version)):
            raise MLSModelError("model_name should follow naming rule. MUST be in alphabet, number, underscore")

        self.models = None
        self.model_name = model_name
        self.model_version = model_version
        self.model_lib = None
        self.model_lib_version = None
        self.features = None

    def save(self, env: MLSENV = None, runtime_env: MLSRuntimeENV = None, force: bool = False) -> None:
        """
        모델 바이너리(model.joblib)와 정보(model.json)를 MLS 모델 레지스트리에 등록합니다.

        `sktmls.ModelRegistry.save`와 동일하게 동작합니다.

        ## Args

        - env: (`sktmls.MLSENV`) 접근할 MLS 환경 (`sktmls.MLSENV.DEV`|`sktmls.MLSENV.STG`|`sktmls.MLSENV.PRD`) (기본값: `sktmls.MLSENV.STG`)
        - runtime_env: (`sktmls.MLSRuntimeENV`) 클라이언트가 실행되는 환경 (`sktmls.MLSRuntimeENV.YE`|`sktmls.MLSRuntimeENV.EDD`|`sktmls.MLSRuntimeENV.LOCAL`) (기본값: `sktmls.MLSRuntimeENV.LOCAL`)
        - force: (bool) 이미 모델 레지스트리에 등록된 경우 덮어 쓸 것인지 여부 (기본값: `False`)
        """
        model_registry = ModelRegistry(env=env, runtime_env=runtime_env)
        model_registry.save(self, force)

    @abstractmethod
    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        pass


class MLSLightGBMModel(MLSModel):
    """
    MLS 모델 레지스트리에 등록되는 단일 LightGBM 기반 상위 클래스입니다.
    """

    def __init__(self, model, model_name: str, model_version: str, features: List[str] = None):
        super().__init__(model_name, model_version)
        self.models = [model]
        self.model_lib = "lightgbm"
        self.model_lib_version = lightgbm.__version__

        if features:
            self.features = features
        else:
            raise MLSModelError("`features`가 없습니다.")


class MLSXGBoostModel(MLSModel):
    """
    MLS 모델 레지스트리에 등록되는 단일 XGBoost 기반 상위 클래스입니다.
    """

    def __init__(self, model, model_name: str, model_version: str, features: List[str]):
        super().__init__(model_name, model_version)
        self.models = [model]
        self.model_lib = "xgboost"
        self.model_lib_version = xgboost.__version__

        if features:
            self.features = features
        else:
            raise MLSModelError("`features`가 없습니다.")


class MLSCatBoostModel(MLSModel):
    """
    MLS 모델 레지스트리에 등록되는 단일 CatBoost 기반 상위 클래스입니다.
    """

    def __init__(self, model, model_name: str, model_version: str, features: List[str]):
        super().__init__(model_name, model_version)
        self.models = [model]
        self.model_lib = "catboost"
        self.model_lib_version = catboost.__version__

        if features:
            self.features = features
        else:
            raise MLSModelError("`features`가 없습니다.")


class MLSRuleModel(MLSModel):
    """
    MLS 모델 레지스트리에 등록되는 Rule 기반 상위 클래스입니다.
    """

    def __init__(self, model_name: str, model_version: str, features: List[str]):
        super().__init__(model_name, model_version)
        self.model_lib = "rule"
        self.model_lib_version = "N/A"

        if features:
            self.features = features
        else:
            raise MLSModelError("`features`가 없습니다.")


class MLSGenericModel(MLSModel):
    """
    MLS 모델 레지스트리에 등록되는 제네릭 모델 기반 상위 클래스입니다.
    """

    def __init__(self, models, model_name: str, model_version: str, features: List[str]):
        assert type(models) == list, "`models`는 list 타입이어야 합니다."

        super().__init__(model_name, model_version)
        self.models = models
        self.model_lib = "etc"
        self.model_lib_version = "N/A"

        if features:
            self.features = features
        else:
            raise MLSModelError("`features`가 없습니다.")

    def set_model_lib(self, model_lib: str):
        """
        모델이 사용한 라이브러리 정보를 세팅하기 위한 함수입니다.
        """
        self.model_lib = model_lib
        if model_lib == "lightgbm":
            self.model_lib_version = lightgbm.__version__
        elif model_lib == "catboost":
            self.model_lib_version = catboost.__version__
        elif model_lib == "xgboost":
            self.model_lib_version = xgboost.__version__
        elif model_lib == "autogluon":
            self.model_lib_version == autogluon.__version__
        elif model_lib == "etc" or model_lib == "rule":
            self.model_lib_version == "N/A"



from typing import Any, Dict, List

from sktmls.models import MLSModelError, MLSRuleModel


class InfoDefectRuleModel(MLSRuleModel):
    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        if len(self.features) != len(x):
            raise MLSModelError("The length of input is different from that of features in model.json")

        svc_scrb_period = x[0] if x[0] is not None else -1
        eqp_chg_period = x[1] if x[1] is not None else 99999

        if (svc_scrb_period <= 30) or (eqp_chg_period >= 1 and eqp_chg_period <= 30):
            return {"items": [{"id": "INF0000004", "name": "초기 불량 대처 방법 안내", "props": {}, "type": "info_defect"}]}
        else:
            return {"items": []}



import pytest

import numpy as np

import lightgbm

from sktmls.meta_tables import MetaTable
from sktmls.models.contrib import LightGBMDeviceModel


class TestLightGBMDeviceModels:
    @pytest.fixture(scope="class")
    def param(self):
        return {
            "model": LightGBMDeviceModel(
                model=lightgbm.basic.Booster(train_set=lightgbm.Dataset(np.array([[1, 2]]), np.array([1]))),
                model_name="test_model",
                model_version="test_version",
                model_features=["eqp_mdl_cd", "age"] + ["feature1", "feature2", "feature3", "feature4"],
                default_model_feature_values=[0.0, "N", 10.0, "S"],
                label_indices={
                    "삼성전자_120만원": 0,
                    "삼성전자_160만원": 1,
                    "Apple_120만원": 2,
                    "Apple_160만원": 3,
                    "기타_120만원": 4,
                    "기타_160만원": 5,
                },
                product_classes={
                    "Apple_120만원": [
                        {
                            "product_grp_id": "000004352",
                            "product_grp_nm": "iPhone SE (2020)",
                            "mfact_nm": "Apple",
                            "eqp_mdl_cd": "A23F",
                            "rep_eqp_mdl_cd": "A23F",
                            "rep_eqp_yn": "Y",
                            "color_hex": "1F2120",
                            "color_nm": "블랙",
                            "network_type": "4G",
                            "device_weight": 0.02356698,
                        },
                        {
                            "product_grp_id": "000003972",
                            "product_grp_nm": "iPhone 11",
                            "mfact_nm": "Apple",
                            "eqp_mdl_cd": "A1G6",
                            "rep_eqp_mdl_cd": "A1G6",
                            "rep_eqp_yn": "Y",
                            "color_hex": "1F2120",
                            "color_nm": "블랙",
                            "network_type": "4G",
                            "device_weight": 0.01638297,
                        },
                    ],
                    "Apple_160만원": [
                        {
                            "product_grp_id": "000003973",
                            "product_grp_nm": "iPhone 11 Pro",
                            "mfact_nm": "Apple",
                            "eqp_mdl_cd": "A1GR",
                            "rep_eqp_mdl_cd": "A1GR",
                            "rep_eqp_yn": "Y",
                            "color_hex": "52514F",
                            "color_nm": "스페이스 그레이",
                            "network_type": "4G",
                            "device_weight": 0.00767129,
                        }
                    ],
                    "삼성전자_120만원": [
                        {
                            "product_grp_id": "000004552",
                            "product_grp_nm": "갤럭시 노트20 5G",
                            "mfact_nm": "삼성전자",
                            "eqp_mdl_cd": "A26X",
                            "rep_eqp_mdl_cd": "A26X",
                            "rep_eqp_yn": "Y",
                            "color_hex": "4A4A4A",
                            "color_nm": "미스틱 그레이",
                            "network_type": "5G",
                            "device_weight": 0.22045896,
                        },
                        {
                            "product_grp_id": "000004432",
                            "product_grp_nm": "갤럭시 A 퀀텀",
                            "mfact_nm": "삼성전자",
                            "eqp_mdl_cd": "A1XG",
                            "rep_eqp_mdl_cd": "A1XG",
                            "rep_eqp_yn": "Y",
                            "color_hex": "191C25",
                            "color_nm": "프리즘 큐브 블랙",
                            "network_type": "5G",
                            "device_weight": 0.07969578,
                        },
                        {
                            "product_grp_id": "000003132",
                            "product_grp_nm": "갤럭시 노트9 (재출시)",
                            "mfact_nm": "삼성전자",
                            "eqp_mdl_cd": "A0M0",
                            "rep_eqp_mdl_cd": "A0M0",
                            "rep_eqp_yn": "Y",
                            "color_hex": "000000",
                            "color_nm": "미드나잇블랙",
                            "network_type": "4G",
                            "device_weight": 0.04440128,
                        },
                    ],
                    "삼성전자_160만원": [
                        {
                            "product_grp_id": "000004553",
                            "product_grp_nm": "갤럭시 노트20 Ultra 5G",
                            "mfact_nm": "삼성전자",
                            "eqp_mdl_cd": "A26Y",
                            "rep_eqp_mdl_cd": "A26Y",
                            "rep_eqp_yn": "Y",
                            "color_hex": "1F1F1F",
                            "color_nm": "미스틱 블랙",
                            "network_type": "5G",
                            "device_weight": 0.14155994,
                        },
                    ],
                },
                products={
                    "000001153": {
                        "name": "갤럭시 S7 엣지",
                        "type": "device",
                        "context_features": ["context_popular_apple_device"],
                        "random_context": False,
                    },
                    "000001672": {
                        "name": "iPhone 7",
                        "type": "device",
                        "context_features": ["context_popular_apple_device"],
                        "random_context": False,
                    },
                    "000004352": {
                        "name": "iPhone SE 2020",
                        "type": "device",
                        "context_features": [
                            "context_age10",
                            "context_age20",
                            "context_age30",
                            "context_default",
                            "context_popular_apple_device",
                        ],
                        "random_context": False,
                    },
                    "000003972": {
                        "name": "iPhone 11",
                        "type": "device",
                        "context_features": ["context_popular_apple_device"],
                        "random_context": False,
                    },
                    "000004552": {
                        "name": "갤럭시 노트20 5G",
                        "type": "device",
                        "context_features": [
                            "context_popular_device",
                            "context_popular_samsung_device",
                            "context_age20",
                            "context_age10",
                            "context_age50",
                            "context_age40",
                            "context_age60plus",
                            "context_age30",
                        ],
                        "random_context": False,
                    },
                    "000004553": {
                        "name": "갤럭시 노트20 Ultra 5G",
                        "type": "device",
                        "context_features": [
                            "context_popular_samsung_device",
                            "context_age20",
                            "context_age50",
                            "context_age40",
                            "context_age30",
                        ],
                        "random_context": False,
                    },
                },
                device_meta=MetaTable(
                    **{
                        "id": 1,
                        "name": "name",
                        "description": "desc",
                        "schema": {},
                        "items": [
                            {
                                "name": "A1GW",
                                "values": {
                                    "color_nm": "실버",
                                    "mfact_nm": "Apple",
                                    "color_hex": "EBEBE3",
                                    "rep_eqp_yn": "N",
                                    "network_type": "4G",
                                    "product_grp_id": "000003973",
                                    "product_grp_nm": "iPhone 11 Pro",
                                    "rep_eqp_mdl_cd": "A1GV",
                                },
                            },
                            {
                                "name": "A26Y",
                                "values": {
                                    "color_nm": "미스틱 블랙",
                                    "mfact_nm": "삼성전자",
                                    "color_hex": "1F1F1F",
                                    "rep_eqp_yn": "Y",
                                    "network_type": "5G",
                                    "product_grp_id": "000004553",
                                    "product_grp_nm": "갤럭시 노트20 Ultra 5G",
                                    "rep_eqp_mdl_cd": "A26Y",
                                },
                            },
                        ],
                        "user": "user",
                        "created_at": "1",
                        "updated_at": "1",
                    }
                ),
                num_pick=3,
            )
        }

    def test_000_generic_context_apple(self, param, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([[0.1, 0.2, 0.3, 0.4]]))
        assert len(param.get("model").predict(["A1GW", 25] + [0.0, "N", 10.0, "S"]).get("items")) == 3
        assert param.get("model").predict(["A1GW", 25] + [0.0, "N", 10.0, "S"]) == {
            "items": [
                {
                    "id": "000004352",
                    "name": "iPhone SE (2020)",
                    "props": {"context_id": "context_age20"},
                    "type": "device",
                },
                {
                    "id": "000003972",
                    "name": "iPhone 11",
                    "props": {"context_id": "context_popular_apple_device"},
                    "type": "device",
                },
                {
                    "id": "000004553",
                    "name": "갤럭시 노트20 Ultra 5G",
                    "props": {"context_id": "context_popular_samsung_device"},
                    "type": "device",
                },
            ]
        }

    def test_001_generic_context_samsung(self, param, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([[0.1, 0.2, 0.3, 0.4]]))
        assert len(param.get("model").predict(["A26Y", 32] + [0.0, "N", 9.0, "S"]).get("items")) == 3
        assert param.get("model").predict(["A26Y", 32] + [0.0, "N", 9.0, "S"]) == {
            "items": [
                {
                    "id": "000004552",
                    "name": "갤럭시 노트20 5G",
                    "props": {"context_id": "context_popular_device"},
                    "type": "device",
                },
                {"id": "000004432", "name": "갤럭시 A 퀀텀", "props": {"context_id": "context_default"}, "type": "device"},
                {
                    "id": "000004352",
                    "name": "iPhone SE (2020)",
                    "props": {"context_id": "context_age30"},
                    "type": "device",
                },
            ]
        }
